/* ====================================================================
 * Copyright (c) 1995 The Apache Group.  All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in
 *    the documentation and/or other materials provided with the
 *    distribution.
 *
 * 3. All advertising materials mentioning features or use of this
 *    software must display the following acknowledgment:
 *    "This product includes software developed by the Apache Group
 *    for use in the Apache HTTP server project (http://www.apache.org/)."
 *
 * 4. The names "Apache Server" and "Apache Group" must not be used to
 *    endorse or promote products derived from this software without
 *    prior written permission.
 *
 * 5. Redistributions of any form whatsoever must retain the following
 *    acknowledgment:
 *    "This product includes software developed by the Apache Group
 *    for use in the Apache HTTP server project (http://www.apache.org/)."
 *
 * THIS SOFTWARE IS PROVIDED BY THE APACHE GROUP ``AS IS'' AND ANY
 * EXPRESSED OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL THE APACHE GROUP OR
 * IT'S CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 * NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
 * STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED
 * OF THE POSSIBILITY OF SUCH DAMAGE.
 * ====================================================================
 *
 * This software consists of voluntary contributions made by many
 * individuals on behalf of the Apache Group and was originally based
 * on public domain software written at the National Center for
 * Supercomputing Applications, University of Illinois, Urbana-Champaign.
 * For more information on the Apache Group and the Apache HTTP server
 * project, please see <http://www.apache.org/>.
 *
 */


/*
 * http_auth_msql: authentication
 *
 * Rob McCool & Brian Behlendorf.
 *
 * Adapted to Shambhala by rst.
 *
 * converted to use MySQL by Vivek Khera <khera@kciLink.com>
 *
 * now maintained by Sam Brauer <sbrauer@users.sourceforge.net>
 * 2002-08-09: ported to Apache-2.x
 * 2002-09-06: merged Apache1 and Apache2 code together with #ifdef's
 * 2002-09-08: merged in some contributed patches:
 * - added AuthMySQLUserCondition and AuthMySQLGroupCondition,
 * based on a patch from Lukas Gradl <l.gradl@business-solutions.at>
 * - added AuthMySQLScrambledPasswords based on patch from 
 * Chris Hamilton <chris@turlyming.com>
 * 2002-09-12: released on sourceforge.net
 * 2002-09-16: released on sourceforge.net (after change to README)
 * 2003-03-16:
 * - added Artur Kedzierski's mysql 4 change to
 * use mysql_real_connect().
 * - added Martin Willner's patch for MD5 support (enable by defining
 * MD5_ENABLED).
 * - moved usage notes to README file and updated a bit.
 * 2003-03-19: added Iain Patterson's patch to call mysql_init() before mysql_real_connect().
 * 2003-05-10: removed the old MD5 stuff that depended on openssl and replaced with a call to apache's ap_md5() function.
 */

/*
 * Module definition information - the part between the -START and -END
 * lines below is used by Configure. This could be stored in a separate
 * instead.
 *
 * MODULE-DEFINITION-START
 * Name: mysql_auth_module
 * ConfigStart
     MYSQL_LIB="-L/usr/local/lib/mysql -lmysqlclient -lm"
     if [ "X$MYSQL_LIB" != "X" ]; then
         LIBS="$LIBS $MYSQL_LIB"
         echo " + using $MYSQL_LIB for MySQL support"
     fi
 * ConfigEnd
 * MODULE-DEFINITION-END
 */


#ifdef APACHE2
#define PCALLOC apr_pcalloc
#define SNPRINTF apr_snprintf
#define PSTRDUP apr_pstrdup
#else
#define PCALLOC ap_pcalloc
#define SNPRINTF ap_snprintf
#define PSTRDUP ap_pstrdup
#endif

#include "httpd.h"
#include "http_config.h"
#include "http_core.h"
#include "http_log.h"
#include "http_protocol.h"
#ifdef APACHE2
#include "http_request.h"   /* for ap_hook_(check_user_id | auth_checker)*/
#endif
#include <mysql/mysql.h>

/*
 * structure to hold the configuration details for the request
 */
typedef struct  {
  char *mysqlhost;		/* host name of db server */
  char *mysqluser;		/* user ID to connect to db server */
  char *mysqlpasswd;		/* password to connect to db server */
  char *mysqlDB;		/* DB name */
  char *mysqlpwtable;		/* user password table */
  char *mysqlgrptable;		/* user group table */
  char *mysqlNameField;		/* field in password/grp table with username */
  char *mysqlPasswordField;	/* field in password table with password */
  char *mysqlGroupField;	/* field in group table with group name */
  int  mysqlCrypted;		/* are passwords encrypted? */
  int  mysqlScrambled;		/* are passwords scrambled? */
  int  mysqlMD5;            /* are passwords MD5? */
  int  mysqlKeepAlive;		/* keep connection persistent? */
  int  mysqlAuthoritative;	/* are we authoritative? */
  int  mysqlNoPasswd;		/* do we ignore password? */
  char *mysqlUserCondition; /* Condition to add to the user where-clause in select query */
  char *mysqlGroupCondition; /* Condition to add to the group where-clause in select query */
} mysql_auth_config_rec;



/*
 * Global handle to db.  If not null, assume it is still valid.
 * MySQL in recent incarnations will re-connect automatically if the
 * connection is closed, so we don't worry about that here.
 */
static MYSQL *mysql_handle = NULL;

/*
 * Callback to close mysql handle when necessary.  Also called when a
 * child httpd process is terminated.
 */
#ifdef APACHE2
static apr_status_t
#else
static void
#endif
mod_auth_mysql_cleanup (void *notused)
{
  if (mysql_handle) mysql_close(mysql_handle);
  mysql_handle = NULL;		/* make sure we don't try to use it later */
#ifdef APACHE2
  return 0;
#endif
}

/*
 * empty function necessary because register_cleanup requires it as one
 * of its parameters
 */
#ifdef APACHE2
static apr_status_t
#else
static void
#endif
mod_auth_mysql_cleanup_child (void *notused)
{
  /* nothing */
#ifdef APACHE2
  return 0;
#endif
}


#ifdef APACHE1
/*
 * handler to do cleanup on child exit
 */
static void
child_exit(server_rec *s, pool *p)
{
  mod_auth_mysql_cleanup(NULL);
}
#endif



/*
 * open connection to DB server if necessary.  Return TRUE if connection
 * is good, FALSE if not able to connect.  If false returned, reason
 * for failure has been logged to error_log file already.
 */
#ifndef TRUE
#define TRUE 1
#endif
#ifndef FALSE
#define FALSE 0
#endif

static int
open_db_handle(request_rec *r, mysql_auth_config_rec *m)
{
  static MYSQL mysql_conn;
  char *db_host;

  if (mysql_handle) return TRUE; /* already open */

  mysql_handle = mysql_init(&mysql_conn);
  if (! mysql_handle) {
#ifdef APACHE2
    ap_log_rerror (APLOG_MARK, APLOG_ERR, 0, r,
    "MySQL ERROR: %s", mysql_error(&mysql_conn));
#else
    ap_log_error (APLOG_MARK, APLOG_ERR, r->server,
    "MySQL error: %s", mysql_error(&mysql_conn));
#endif
  }

  if (!m->mysqlhost || strcmp(m->mysqlhost,"localhost") == 0) {
    db_host = NULL;
  } else {
    db_host = m->mysqlhost;
  }

  mysql_handle=mysql_real_connect(&mysql_conn,db_host,m->mysqluser,m->mysqlpasswd, NULL, 0, NULL, 0);

  if (mysql_handle) {

    if (!m->mysqlKeepAlive) {
      /* close when request done */
#ifdef APACHE2
	  apr_pool_cleanup_register
#else
	  ap_register_cleanup
#endif
			  (r->pool, (void *)NULL,
			  mod_auth_mysql_cleanup,
			  mod_auth_mysql_cleanup_child);
    } /* ELSE...
       * Child process is notified when it is terminated so we
       * do a graceful close to the server in that handler.
       */

  } else {			/* failed to get MySQL connection */
#ifdef APACHE2
	ap_log_rerror (APLOG_MARK, APLOG_ERR, 0, r,
			"MySQL ERROR: %s", mysql_error(&mysql_conn));
#else
    ap_log_error (APLOG_MARK, APLOG_ERR, r->server,
			"MySQL error: %s", mysql_error(&mysql_conn));
#endif

    return FALSE;
  }

  return TRUE;
}


#ifdef APACHE2
static void *
create_mysql_auth_dir_config (apr_pool_t *p, char *d)
#else
static void *
create_mysql_auth_dir_config (pool *p, char *d)
#endif
{
  mysql_auth_config_rec *m = PCALLOC(p, sizeof(mysql_auth_config_rec));
  if (!m) return NULL;		/* failure to get memory is a bad thing */

  /* default values */
  m->mysqlNameField = "user_name";
  m->mysqlPasswordField = "user_passwd";
  m->mysqlCrypted = 1;		/* passwords are encrypted */

  m->mysqlMD5 = 0;
  m->mysqlScrambled = 0;		/* passwords are scrambled */
  m->mysqlKeepAlive = 0;	/* do not keep persistent connection */
  m->mysqlAuthoritative = 1;	/* we are authoritative source for users */
  m->mysqlNoPasswd = 0;		/* we require password */
  return (void *)m;
}

#ifdef APACHE2
static
command_rec mysql_auth_cmds[] = {
	AP_INIT_TAKE1("AuthMySQLHost", ap_set_string_slot,
	(void *) APR_XtOffsetOf(mysql_auth_config_rec, mysqlhost),
	OR_AUTHCFG, "mysql server host name"),

	AP_INIT_TAKE1("AuthMySQLUser", ap_set_string_slot,
	(void *) APR_XtOffsetOf(mysql_auth_config_rec, mysqluser),
	OR_AUTHCFG, "mysql server user name"),

	AP_INIT_TAKE1("AuthMySQLPassword", ap_set_string_slot,
	(void *) APR_XtOffsetOf(mysql_auth_config_rec, mysqlpasswd),
	OR_AUTHCFG, "mysql server user password"),

	AP_INIT_TAKE1("AuthMySQLDB", ap_set_string_slot,
	(void *) APR_XtOffsetOf(mysql_auth_config_rec, mysqlDB),
	OR_AUTHCFG, "mysql database name"),

	AP_INIT_TAKE1("AuthMySQLUserTable", ap_set_string_slot,
	(void *) APR_XtOffsetOf(mysql_auth_config_rec, mysqlpwtable),
	OR_AUTHCFG, "mysql user table name"),

	AP_INIT_TAKE1("AuthMySQLGroupTable", ap_set_string_slot,
	(void *) APR_XtOffsetOf(mysql_auth_config_rec, mysqlgrptable),
	OR_AUTHCFG, "mysql group table name"),

	AP_INIT_TAKE1("AuthMySQLNameField", ap_set_string_slot,
	(void *) APR_XtOffsetOf(mysql_auth_config_rec, mysqlNameField),
	OR_AUTHCFG, "mysql User ID field name within table"),

	AP_INIT_TAKE1("AuthMySQLGroupField", ap_set_string_slot,
	(void *) APR_XtOffsetOf(mysql_auth_config_rec, mysqlGroupField),
	OR_AUTHCFG, "mysql Group field name within table"),

	AP_INIT_TAKE1("AuthMySQLPasswordField", ap_set_string_slot,
	(void *) APR_XtOffsetOf(mysql_auth_config_rec, mysqlPasswordField),
	OR_AUTHCFG, "mysql Password field name within table"),

	AP_INIT_FLAG("AuthMySQLCryptedPasswords", ap_set_flag_slot,
	(void *) APR_XtOffsetOf(mysql_auth_config_rec, mysqlCrypted),
	OR_AUTHCFG, "mysql passwords are stored encrypted if On"),

	AP_INIT_FLAG("AuthMySQLScrambledPasswords", ap_set_flag_slot,
	(void *) APR_XtOffsetOf(mysql_auth_config_rec, mysqlScrambled),
	OR_AUTHCFG, "mysql passwords are stored scrambled if On"),

	AP_INIT_FLAG("AuthMySQLMD5Passwords", ap_set_flag_slot,
	(void *) APR_XtOffsetOf(mysql_auth_config_rec, mysqlMD5),
	OR_AUTHCFG, "mysql passwords are stored as MD5 if On"),

	AP_INIT_FLAG("AuthMySQLKeepAlive", ap_set_flag_slot,
	(void *) APR_XtOffsetOf(mysql_auth_config_rec, mysqlKeepAlive),
	OR_AUTHCFG, "mysql connection kept open across requests if On"),

	AP_INIT_FLAG("AuthMySQLAuthoritative", ap_set_flag_slot,
	(void *) APR_XtOffsetOf(mysql_auth_config_rec, mysqlAuthoritative),
	OR_AUTHCFG, "mysql lookup is authoritative if On"),

	AP_INIT_FLAG("AuthMySQLNoPasswd", ap_set_flag_slot,
	(void *) APR_XtOffsetOf(mysql_auth_config_rec, mysqlNoPasswd),
	OR_AUTHCFG, "If On, only check if user exists; ignore password"),

	AP_INIT_TAKE1("AuthMySQLUserCondition", ap_set_string_slot,
	(void *) APR_XtOffsetOf(mysql_auth_config_rec, mysqlUserCondition),
	OR_AUTHCFG, "condition to add to user where-clause"),

	AP_INIT_TAKE1("AuthMySQLGroupCondition", ap_set_string_slot,
	(void *) APR_XtOffsetOf(mysql_auth_config_rec, mysqlGroupCondition),
	OR_AUTHCFG, "condition to add to group where-clause"),

  { NULL }
};
#else
static
command_rec mysql_auth_cmds[] = {
  { "AuthMySQLHost", ap_set_string_slot,
    (void*)XtOffsetOf(mysql_auth_config_rec, mysqlhost),
    OR_AUTHCFG, TAKE1, "mysql server host name" },

  { "AuthMySQLUser", ap_set_string_slot,
    (void*)XtOffsetOf(mysql_auth_config_rec, mysqluser),
    OR_AUTHCFG, TAKE1, "mysql server user name" },

  { "AuthMySQLPassword", ap_set_string_slot,
    (void*)XtOffsetOf(mysql_auth_config_rec, mysqlpasswd),
    OR_AUTHCFG, TAKE1, "mysql server user password" },

  { "AuthMySQLDB", ap_set_string_slot,
    (void*)XtOffsetOf(mysql_auth_config_rec, mysqlDB),
    OR_AUTHCFG, TAKE1, "mysql database name" },

  { "AuthMySQLUserTable", ap_set_string_slot,
    (void*)XtOffsetOf(mysql_auth_config_rec, mysqlpwtable),
    OR_AUTHCFG, TAKE1, "mysql user table name" },

  { "AuthMySQLGroupTable", ap_set_string_slot,
    (void*)XtOffsetOf(mysql_auth_config_rec, mysqlgrptable),
    OR_AUTHCFG, TAKE1, "mysql group table name" },

  { "AuthMySQLNameField", ap_set_string_slot,
    (void*)XtOffsetOf(mysql_auth_config_rec, mysqlNameField),
    OR_AUTHCFG, TAKE1, "mysql User ID field name within table" },

  { "AuthMySQLGroupField", ap_set_string_slot,
    (void*)XtOffsetOf(mysql_auth_config_rec, mysqlGroupField),
    OR_AUTHCFG, TAKE1, "mysql Group field name within table" },

  { "AuthMySQLPasswordField", ap_set_string_slot,
    (void*)XtOffsetOf(mysql_auth_config_rec, mysqlPasswordField),
    OR_AUTHCFG, TAKE1, "mysql Password field name within table" },

  { "AuthMySQLCryptedPasswords", ap_set_flag_slot,
    (void*)XtOffsetOf(mysql_auth_config_rec, mysqlCrypted),
    OR_AUTHCFG, FLAG, "mysql passwords are stored encrypted if On" },

  { "AuthMySQLScrambledPasswords", ap_set_flag_slot,
    (void*)XtOffsetOf(mysql_auth_config_rec, mysqlScrambled),
    OR_AUTHCFG, FLAG, "mysql passwords are stored scrambled if On" },

  { "AuthMySQLMD5Passwords", ap_set_flag_slot,
    (void*)XtOffsetOf(mysql_auth_config_rec, mysqlMD5),
    OR_AUTHCFG, FLAG, "mysql passwords are stored as MD5 if On" },

  { "AuthMySQLKeepAlive", ap_set_flag_slot,
    (void*)XtOffsetOf(mysql_auth_config_rec, mysqlKeepAlive),
    OR_AUTHCFG, FLAG, "mysql connection kept open across requests if On" },

  { "AuthMySQLAuthoritative", ap_set_flag_slot,
    (void*)XtOffsetOf(mysql_auth_config_rec, mysqlAuthoritative),
    OR_AUTHCFG, FLAG, "mysql lookup is authoritative if On" },

  { "AuthMySQLNoPasswd", ap_set_flag_slot,
    (void*)XtOffsetOf(mysql_auth_config_rec, mysqlNoPasswd),
    OR_AUTHCFG, FLAG, "If On, only check if user exists; ignore password" },

  { "AuthMySQLUserCondition", ap_set_string_slot,
    (void*)XtOffsetOf(mysql_auth_config_rec, mysqlUserCondition),
    OR_AUTHCFG, TAKE1, "condition to add to user where-clause" },

  { "AuthMySQLGroupCondition", ap_set_string_slot,
    (void*)XtOffsetOf(mysql_auth_config_rec, mysqlGroupCondition),
    OR_AUTHCFG, TAKE1, "condition to add to group where-clause" },

  { NULL }
};
#endif

module mysql_auth_module;

/*
 * Fetch and return password string from database for named user.
 * If we are in NoPasswd mode, returns user name instead.
 * If user or password not found, returns NULL
 */
static char *
get_mysql_pw(request_rec *r, char *user, mysql_auth_config_rec *m)
{
  MYSQL_RES *result;
  char *pw = NULL;		/* password retrieved */
  char *sql_safe_user = NULL;
  int ulen;
  char query[MAX_STRING_LEN];

  if(!open_db_handle(r,m)) {
    return NULL;		/* failure reason already logged */
  }

  if (mysql_select_db(mysql_handle,m->mysqlDB) != 0) {
#ifdef APACHE2
	ap_log_rerror (APLOG_MARK, APLOG_ERR, 0, r,
		"MySQL ERROR: %s", mysql_error(mysql_handle));
#else
    ap_log_error (APLOG_MARK, APLOG_ERR, r->server,
		"MySQL error: %s", mysql_error(mysql_handle));
#endif
    return NULL;
  }

  /*
   * If we are not checking for passwords, there may not be a password field
   * in the database.  We just look up the name field value in this case
   * since it is guaranteed to exist.
   */
  if (m->mysqlNoPasswd) {
    m->mysqlPasswordField = m->mysqlNameField;
  }

  ulen = strlen(user);
  sql_safe_user = PCALLOC(r->pool, ulen*2+1);
  mysql_escape_string(sql_safe_user,user,ulen);
  if (m->mysqlUserCondition) {
    SNPRINTF(query,sizeof(query)-1,"SELECT %s FROM %s WHERE %s='%s' AND %s",
	      m->mysqlPasswordField, m->mysqlpwtable,
	      m->mysqlNameField, sql_safe_user, m->mysqlUserCondition);
  } else {
    SNPRINTF(query,sizeof(query)-1,"SELECT %s FROM %s WHERE %s='%s'",
	      m->mysqlPasswordField, m->mysqlpwtable,
	      m->mysqlNameField, sql_safe_user);
  }
  if (mysql_query(mysql_handle, query) != 0) {
#ifdef APACHE2
	ap_log_rerror (APLOG_MARK, APLOG_ERR, 0, r,
			"MySQL ERROR: %s: %s", mysql_error(mysql_handle), r->uri);
#else
    ap_log_error (APLOG_MARK, APLOG_ERR, r->server,
		  "MySQL error %s: %s", mysql_error(mysql_handle),r->uri);
#endif
    return NULL;
  }

  result = mysql_store_result(mysql_handle);
  /*if (result && (mysql_num_rows(result) == 1)) {*/
  if (result && (mysql_num_rows(result) >= 1)) {
    MYSQL_ROW data = mysql_fetch_row(result);
    if (data[0]) {
      pw = (char *) PSTRDUP(r->pool, data[0]);
    } else {		/* no password in mysql table returns NULL */
      /* this should never happen, but test for it anyhow */
#ifdef APACHE2
      ap_log_rerror (APLOG_MARK, APLOG_NOERRNO|APLOG_ERR, 0, r, 
		   "MySQL user %s has no valid password: %s", user, r->uri);
#else
      ap_log_error(APLOG_MARK, APLOG_NOERRNO|APLOG_ERR, r->server,
		   "MySQL user %s has no valid password: %s", user, r->uri);
#endif
      mysql_free_result(result);
      return NULL;
    }
  }

  if (result) mysql_free_result(result);

  return pw;
}

/*
 * get list of groups from database.  Returns array of pointers to strings
 * the last of which is NULL.  returns NULL pointer if user is not member
 * of any groups.
 */
static char **
get_mysql_groups(request_rec *r, char *user, mysql_auth_config_rec *m)
{
  MYSQL_RES *result;
  char **list = NULL;
  char query[MAX_STRING_LEN];
  char *sql_safe_user;
  int ulen;

  if(!open_db_handle(r,m)) {
    return NULL;		/* failure reason already logged */
  }

  if (mysql_select_db(mysql_handle,m->mysqlDB) != 0) {
#ifdef APACHE2
    ap_log_rerror (APLOG_MARK, APLOG_ERR, 0, r, 
		  "MySQL error %s: %s", mysql_error(mysql_handle),r->uri);
#else
    ap_log_error (APLOG_MARK, APLOG_ERR, r->server,
		  "MySQL error %s: %s", mysql_error(mysql_handle),r->uri);
#endif
    return NULL;
  }

  ulen = strlen(user);
  sql_safe_user = PCALLOC(r->pool, ulen*2+1);
  mysql_escape_string(sql_safe_user,user,ulen);
  if (m->mysqlGroupCondition) {
    SNPRINTF(query,sizeof(query)-1,"SELECT %s FROM %s WHERE %s='%s' AND %s",
	      m->mysqlGroupField, m->mysqlgrptable,
	      m->mysqlNameField, sql_safe_user, m->mysqlGroupCondition);
  } else {
    SNPRINTF(query,sizeof(query)-1,"SELECT %s FROM %s WHERE %s='%s'",
	      m->mysqlGroupField, m->mysqlgrptable,
	      m->mysqlNameField, sql_safe_user);
  }
  if (mysql_query(mysql_handle, query) != 0) {
#ifdef APACHE2
    ap_log_rerror (APLOG_MARK, APLOG_ERR, 0, r, 
		  "MySQL error %s: %s", mysql_error(mysql_handle),r->uri);
#else
    ap_log_error (APLOG_MARK, APLOG_ERR, r->server,
		  "MySQL error %s: %s", mysql_error(mysql_handle),r->uri);
#endif
    return NULL;
  }

  result = mysql_store_result(mysql_handle);
  if (result && (mysql_num_rows(result) > 0)) {
    int i = mysql_num_rows(result);
    list = (char **) PCALLOC(r->pool, sizeof(char *) * (i+1));
    list[i] = NULL;		/* last element in array is NULL */
    while (i--) {		/* populate the array elements */
      MYSQL_ROW data = mysql_fetch_row(result);
      if (data[0])
	list[i] = (char *) PSTRDUP(r->pool, data[0]);
      else
	list[i] = "";		/* if no data, make it empty, not NULL */
    }
  }

  if (result) mysql_free_result(result);

  return list;
}

/*
 * callback from Apache to do the authentication of the user to his
 * password.
 */
static int
mysql_authenticate_basic_user (request_rec *r)
{
  int passwords_match = 1;
  char *user;
  mysql_auth_config_rec *sec =
    (mysql_auth_config_rec *)ap_get_module_config (r->per_dir_config,
						   &mysql_auth_module);
  conn_rec *c = r->connection;
  const char *sent_pw, *real_pw;
  int res;
  char *scrambled_sent_pw;

  if ((res = ap_get_basic_auth_pw (r, &sent_pw)))
    return res;

  if(!sec->mysqlpwtable)	/* not configured for mysql authorization */
    return DECLINED;

#ifdef APACHE2
  user = r->user;
#else
  user = c->user;
#endif
  if(!(real_pw = get_mysql_pw(r, user, sec)))
  {
    /* user not found in database */
    if (!sec->mysqlAuthoritative)
      return DECLINED;		/* let other schemes find user */

#ifdef APACHE2
    ap_log_rerror (APLOG_MARK, APLOG_NOERRNO|APLOG_ERR, 0, r, 
		 "MySQL user %s not found: %s", user, r->uri);
#else
    ap_log_error(APLOG_MARK, APLOG_NOERRNO|APLOG_ERR, r->server,
		 "MySQL user %s not found: %s", user, r->uri);
#endif
    ap_note_basic_auth_failure (r);
#ifdef APACHE2
    return HTTP_UNAUTHORIZED;
#else
	return AUTH_REQUIRED;
#endif
  }

  /* if we don't require password, just return ok since they exist */
  if (sec->mysqlNoPasswd) {
    return OK;
  }

  /* if AuthMySQLScrambledPassword is On, compare the scrambled password */
  if(sec->mysqlScrambled) {
    scrambled_sent_pw = PCALLOC(r->pool, 33);
    make_scrambled_password(scrambled_sent_pw, sent_pw);
    if(strcmp(real_pw, scrambled_sent_pw)) {
	  passwords_match = 0;
    }
  } else if(sec->mysqlMD5) {
    if(strcmp(real_pw,ap_md5(r->pool, sent_pw))) {
      passwords_match = 0;
    }
  } else {
    /* compare the password, possibly encrypted */
    if(strcmp(real_pw, sec->mysqlCrypted ? crypt(sent_pw,real_pw) : sent_pw)) {
	  passwords_match = 0;
	}
  }

  if(passwords_match) {
	  return OK;
  } else {
#ifdef APACHE2
    ap_log_rerror (APLOG_MARK, APLOG_NOERRNO|APLOG_ERR, 0, r, 
		 "user %s: password mismatch: %s", user, r->uri);
#else
    ap_log_error(APLOG_MARK, APLOG_NOERRNO|APLOG_ERR, r->server,
		 "user %s: password mismatch: %s", user, r->uri);
#endif
    ap_note_basic_auth_failure (r);
#ifdef APACHE2
    return HTTP_UNAUTHORIZED;
#else
	return AUTH_REQUIRED;
#endif
  }
}

/*
 * check if user is member of at least one of the necessary group(s)
 */
static int
mysql_check_auth(request_rec *r)
{
  mysql_auth_config_rec *sec =
    (mysql_auth_config_rec *)ap_get_module_config(r->per_dir_config,
						  &mysql_auth_module);
#ifdef APACHE2
  char *user = r->user;
#else
  char *user = r->connection->user;
#endif
  int method = r->method_number;

#ifdef APACHE2
  const apr_array_header_t *reqs_arr = ap_requires(r);
#else
  const array_header *reqs_arr = ap_requires(r);
#endif

  require_line *reqs = reqs_arr ? (require_line *)reqs_arr->elts : NULL;

  register int x;
  char **groups = NULL;

  if (!sec->mysqlGroupField) return DECLINED; /* not doing groups here */
  if (!reqs_arr) return DECLINED; /* no "require" line in access config */

  /* if the group table is not specified, use the same as for password */
  if (!sec->mysqlgrptable) sec->mysqlgrptable = sec->mysqlpwtable;

  for(x = 0; x < reqs_arr->nelts; x++) {
    const char *t, *want;

    if (!(reqs[x].method_mask & (1 << method))) continue;

    t = reqs[x].requirement;
    want = ap_getword(r->pool, &t, ' ');

    if(!strcmp(want,"group")) {
      /* check for list of groups from database only first time thru */
      if (!groups && !(groups = get_mysql_groups(r, user, sec))) {
#ifdef APACHE2
    ap_log_rerror (APLOG_MARK, APLOG_NOERRNO|APLOG_ERR, 0, r, 
		     "mysql user %s not in group table %s: %s",
		     user, sec->mysqlgrptable, r->uri);
#else
	ap_log_error(APLOG_MARK, APLOG_NOERRNO|APLOG_ERR, r->server,
		     "mysql user %s not in group table %s: %s",
		     user, sec->mysqlgrptable, r->uri);
#endif
	ap_note_basic_auth_failure(r);
#ifdef APACHE2
	return HTTP_UNAUTHORIZED;
#else
	return AUTH_REQUIRED;
#endif
      }

      /* loop through list of groups specified in htaccess file */
      while(t[0]) {
	int i = 0;
	want = ap_getword(r->pool, &t, ' ');
	/* compare against each group to which this user belongs */
	while(groups[i]) {	/* last element is NULL */
	  if(!strcmp(groups[i],want))
	    return OK;		/* we found the user! */
	  ++i;
	}
      }
#ifdef APACHE2
      ap_log_rerror (APLOG_MARK, APLOG_NOERRNO|APLOG_ERR, 0, r, 
		   "mysql user %s not in right group: %s",user,r->uri);
#else
      ap_log_error(APLOG_MARK, APLOG_NOERRNO|APLOG_ERR, r->server,
		   "mysql user %s not in right group: %s",user,r->uri);
#endif
      ap_note_basic_auth_failure(r);
#ifdef APACHE2
	return HTTP_UNAUTHORIZED;
#else
	return AUTH_REQUIRED;
#endif
    }
  }

  return DECLINED;
}

#ifdef APACHE2
static void register_hooks(apr_pool_t *p)
{
	ap_hook_check_user_id(mysql_authenticate_basic_user, NULL, NULL, APR_HOOK_MIDDLE);
	ap_hook_auth_checker(mysql_check_auth, NULL, NULL, APR_HOOK_MIDDLE);
}
#endif

#ifdef APACHE2
module AP_MODULE_DECLARE_DATA mysql_auth_module =
{
STANDARD20_MODULE_STUFF,
create_mysql_auth_dir_config, /* dir config creater */
NULL,                       /* dir merger --- default is to override */
NULL,                       /* server config */
NULL,                       /* merge server config */
mysql_auth_cmds,              /* command apr_table_t */
register_hooks              /* register hooks */
};
#else
module mysql_auth_module = {
   STANDARD_MODULE_STUFF,
   NULL,			/* initializer */
   create_mysql_auth_dir_config, /* dir config creater */
   NULL,			/* dir merger --- default is to override */
   NULL,			/* server config */
   NULL,			/* merge server config */
   mysql_auth_cmds,		/* command table */
   NULL,			/* handlers */
   NULL,			/* filename translation */
   mysql_authenticate_basic_user, /* check_user_id */
   mysql_check_auth,		/* check auth */
   NULL,			/* check access */
   NULL,			/* type_checker */
   NULL,			/* fixups */
   NULL,			/* logger */
   NULL,			/* header parser */
   NULL,			/* child_init */
   child_exit,			/* child_exit */
   NULL				/* post read-request */
};
#endif

